/*

Reverse Geocache Box
  This program unlocks a box that has reached a certain location.

The device uses the following products from Adafruit Industries:
  Arduino Uno: https://www.adafruit.com/products/50
  Ultimate GPS (version3): http://www.adafruit.com/products/746
  16x2 Character LCD: https://www.adafruit.com/products/181
  TPro Micro Servo SG90: https://www.adafruit.com/products/169
  Half Sized Perma proto: https://www.adafruit.com/products/571

Tutorials for these products found on learn.adafruit.com

Credits:
  NeoGPS version without String by /dev

  Thanks to bnordlund9 for much of the code. This is simplified version
  of his Geobox found here:
      http://www.youtube.com/watch?v=g0060tcuofg

  Engagement Box by Kenton Harris 11/12/2012
  
  Reverse Geocache idea by Mikal Hart of http://arduiniana.org/

*/

#include <math.h>
#include <LiquidCrystal.h>
//#include <NewLiquidCrystal_emul.h>
#include "NMEAGPS.h"
#include "DMS.h"
#include <SoftwareSerial.h>
SoftwareSerial gps_port(3,2);

using namespace NeoGPS;
NMEAGPS gps;

//Servo
#include <PWMServo.h>
PWMServo servoLatch;

//Declarations
const float CLOSE_ENOUGH = 300.0; // miles.  Original was 20

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(7, 8, 6, 10, 11, 12);

const int  ledFix      = 4;
const int  servoPin    = 9;
const int  servoLock   = 180;                  // angle (deg) of "locked" servo
const int  servoUnlock = 105;                  // angle (deg) of "unlocked" servo

// The target location, in degrees * 10,000,000
Location_t there( 384602500,  -1098191667L );
               // 38°27'36.2"N 109°49'15.4"W

void setup()
{
  servoLatch.attach(SERVO_PIN_A);
  servoLatch.write(servoLock);
  delay(50);

  lcd.begin(16, 2);

  Serial.begin(9600);
  Serial.println( F("Debug GPS Test:") );

  gps_port.begin(9600);

  // Configure GPS
  gps_port.print
    ( F( "$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29\r\n" // RMC only...
         "$PMTK220,1000*1F\r\n" ) );    // ...and 1 Hz update rate
}

void loop()
{
  static uint8_t  warningState = 0;
  static uint32_t lastFixTime, lastDotTime;

  // Parse GPS and recalculate RANGE
  while (gps.available( gps_port )) {
    gps_fix fix = gps.read(); // save the latest

    // Set the "fix" LED to on or off
    bool gpsWasFixed = fix.valid.status && (fix.status >= gps_fix::STATUS_STD);
    digitalWrite( ledFix, gpsWasFixed );

    // When we have a location, calculate how far away we are from "there".
    if (fix.valid.location) {
      lastFixTime = millis();

      float range   = fix.location.DistanceMiles   ( there );
      float bearing = fix.location.BearingToDegrees( there );

      //for GPS debug
      Serial.print( F("Here: ") );
      printDMS( fix.location );

      Serial.print( F(" There: ") );
      printDMS( there );

      Serial.print( F("Range: ") );
      Serial.print( range );
      Serial.println( F(" Miles") );

      Serial.print( F("Bearing: ") );
      Serial.println( bearing ); //  degrees CW from North

      // Are we there?
      if (range < CLOSE_ENOUGH) {
        lcd.setCursor(0,1);
        lcd.println( F("Box Unlocking...") );
        delay(10000);
        servoLatch.write(servoUnlock);

        lcd.clear();
        lcd.setCursor(0,1);
        lcd.print( F("Happy Birthday") );

        for (;;); // hang here
      }

      // Nope, just give a little feedback...
      lcd.clear();
      lcd.setCursor(0,0);
      lcd.print( F("Signal Locked") );
      lcd.setCursor(0,1);
      lcd.print( range );
      lcd.print( F(" Miles") );

      warningState = 0; // restart in case we lose the fix later
    }
  }

  // Display warnings when the GPS hasn't had a fix for a while

  if (millis() - lastFixTime > 2000UL) {

    if (warningState == 0) {

      // First, show the warning message...
      lcd.clear();
      lcd.setCursor(0,0);
      lcd.print( F("Acquiring Signal") );

      warningState = 1;

    } else if (warningState < 10) {

      // ...then some dots, two per second...
      if (millis() - lastDotTime > 500UL ) {
        lastDotTime = millis();
        lcd.setCursor( warningState-1, 1 );
        lcd.print( '.' );

        warningState++;
      }

    } else if (warningState == 10) {
      // ... then tell them what to do.
      lcd.setCursor(0,1);
      lcd.println( F("Take box outside") );

      // Don't display anything else until location is valid
      warningState++; // 11
    }
  }
}

void printDMS( const Location_t & loc )
{
  DMS_t dms;
  dms.From( loc.lat() );

  Serial.print( dms.NS() );
  Serial << dms;

  dms.From( loc.lon() );
  Serial.print( dms.EW() );
  Serial << dms;
}

